/* 
 * SMB/Netbios implementation.
 *
 * Version 1.9.
 * SMB parameters and setup
 * Copyright (C) Andrew Tridgell 1992-1995
 * Modified by Jeremy Allison 1995.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#if defined(SMB_PASSWD)

#include "std-defs.h"
#include "loadparm.h"
#include "des.h"
#include "md4.h"
#include "rfcnb.h"
#include "smblib.h"

extern int debug_level;

extern void des_set_odd_parity (des_cblock*);

static __inline void str_to_key (UCHAR *str, UCHAR *key)
{
  int i;

  key[0] = str[0] >> 1;
  key[1] = ((str[0] & 0x01) << 6) | (str[1] >> 2);
  key[2] = ((str[1] & 0x03) << 5) | (str[2] >> 3);
  key[3] = ((str[2] & 0x07) << 4) | (str[3] >> 4);
  key[4] = ((str[3] & 0x0F) << 3) | (str[4] >> 5);
  key[5] = ((str[4] & 0x1F) << 2) | (str[5] >> 6);
  key[6] = ((str[5] & 0x3F) << 1) | (str[6] >> 7);
  key[7] = str[6] & 0x7F;
  for (i = 0; i < 8; i++)
      key[i] = (key[i] << 1);

  des_set_odd_parity ((des_cblock*)key);
}

static __inline void D1 (UCHAR * k, UCHAR *d, UCHAR *out)
{
  des_key_schedule ks;
  des_cblock       deskey;

  str_to_key (k, (UCHAR*)&deskey);
  des_set_key (deskey, ks);
  des_ecb_encrypt (d, out, ks, DES_DECRYPT);
}

static __inline void E1 (UCHAR *k, UCHAR *d, UCHAR *out)
{
  des_key_schedule ks;
  des_cblock       deskey;

  str_to_key (k, (UCHAR*)&deskey);
  des_set_key (deskey, ks);
  des_ecb_encrypt (d, out, ks, DES_ENCRYPT);
}

static __inline void E_P16 (UCHAR *p14, UCHAR *p16)
{
  UCHAR sp7[7];

  /* the following constant makes us compatible with other
   * implementations. Note that publishing this constant does not reduce the
   * security of the encryption mechanism
   */
  UCHAR sp8[] = { 0xAA, 0xD3, 0xB4, 0x35, 0xB5, 0x14, 0x4, 0xEE };
  UCHAR x[8];

  memset (sp7, '\0', sizeof(sp7));

  D1 (sp7, sp8, x);
  E1 (p14, x, p16);
  E1 (p14 + 7, x, p16 + 8);
}

static __inline void E_P24 (UCHAR *p21, UCHAR *c8, UCHAR *p24)
{
  E1 (p21, c8, p24);
  E1 (p21 + 7, c8, p24 + 8);
  E1 (p21 + 14, c8, p24 + 16);
}


/*
 * This implements the X/Open SMB password encryption
 * It takes a password, a 8 byte "crypt key" and puts 24 bytes of 
 * encrypted password into p24
 */
void SMB_Encrypt (const UCHAR *passwd, UCHAR *c8, UCHAR *p24)
{
  UCHAR p14[15], p21[21];

  memset (p21, '\0', sizeof(p21));
  memset (p14, '\0', sizeof(p14)-1);
  StrnCpy ((char*)p14, (char*)passwd, sizeof(p14)-1);

  strupper ((char*)p14);
  E_P16 (p14, p21);
  E_P24 (p21, c8, p24);
}

/*
 * Routines for Windows NT MD4 Hash functions.
 * Return length of WideChar string.
 */
static int _my_wcslen (const UWORD *str)
{
  int len = 0;

  while (*str++ != 0)
    len++;
  return (len);
}

/*
 * Convert a string into an NT UNICODE string.
 * Note that regardless of processor type 
 * this must be in intel (little-endian)
 * format.
 */
static int _my_mbstowcs (UWORD *dst, const UCHAR *src, int len)
{
  UWORD val;
  int   i;

  for (i = 0; i < len; i++)
  {
    val = *src;
    SSVAL (dst, 0, val);
    dst++;
    src++;
    if (val == 0)
      break;
  }
  return (i);
}

/* 
 * Creates the MD4 Hash of the users password in NT UNICODE.
 */
static void E_md4hash (const UCHAR *passwd, UCHAR *p16)
{
  #define PASSWD_LEN  64
  MDstruct MD;
  int      i, len;
  UWORD    wpwd [2*PASSWD_LEN+1];  /* Unicode buffer for passwd */

  /* Password cannot be longer than 128 characters
   */
  len = strlen (passwd);
  if (len > 2*PASSWD_LEN)
      len = 2*PASSWD_LEN;

  /* Password must be converted to NT unicode
   */
  _my_mbstowcs (wpwd, passwd, len);
  wpwd [len] = '\0';

  /* Calculate length in bytes
   */
  len = _my_wcslen (wpwd) * sizeof (UWORD);

  MDbegin (&MD);
  for (i = 0; i + PASSWD_LEN <= len; i += PASSWD_LEN)
      MDupdate (&MD, wpwd + (i / 2), 512);

  MDupdate (&MD, wpwd + (i/2), 8*(len-i));
  SIVAL (p16, 0,  MD.buffer[0]);
  SIVAL (p16, 4,  MD.buffer[1]);
  SIVAL (p16, 8,  MD.buffer[2]);
  SIVAL (p16, 12, MD.buffer[3]);
}

/*
 * Does the NT MD4 hash then DES encryption.
 */
void SMBNT_Encrypt (const UCHAR *passwd, const UCHAR *c8, UCHAR *p24)
{
  UCHAR p21[21];

  memset (p21, '\0', sizeof(p21));
  E_md4hash (passwd, p21);
  E_P24 (p21, c8, p24);
}

#endif /* SMB_PASSWD */

